#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>

#include <windows.h>
#include <commctrl.h>
#include <iphlpapi.h>
#include <psapi.h>

#define UNUSED __attribute__((unused))

#define REGISTRY_KEY "SOFTWARE\\SysTrayUtil"
#define REGISTRY_VALUE_ORDER "Order"

#define ALLOC_FORMAT(name, ...) \
char name[snprintf(NULL, 0, __VA_ARGS__) + 1]; \
sprintf(name, __VA_ARGS__);

static void wperror(const char *text) {
	fprintf(stderr, "%s: %d.\n", text, (int) GetLastError());
}

struct {
	char **sorted;
	char **no_class;
} config = {
	.sorted = NULL,
	.no_class = NULL
};

#define TRAY_ID_CHAR_ANY '*'

struct tray_context_t {
	const struct {
		HWND window;
		UINT id;
		UINT message;
	} *data;
	const wchar_t *tooltip;
	bool allow_send_messages;
	bool retry;
};

static int id_clear_path(char *buffer, int length) {
	int index = -1;
	for (int i = length - 1; i >= 0; i--) {
		if (buffer[i] == '\\') {
			index = i;
			break;
		}
	}
	if (index < 0) {
		return length;
	}
	for (int i = length; i >= index + 1; i--) {
		buffer[i - index - 1] = buffer[i];
	}
	length = length - index - 1;
	if (length >= 4 && buffer[length - 4] == '.' &&
		(buffer[length - 3] == 'e' || buffer[length - 3] == 'E') &&
		(buffer[length - 2] == 'x' || buffer[length - 2] == 'X') &&
		(buffer[length - 1] == 'e' || buffer[length - 1] == 'E')) {
		length -= 4;
		buffer[length] = '\0';
	}
	return length;
}

static int id_simplify(char *buffer, int length, bool lower) {
	for (int i = length - 1; i >= 0; i--) {
		char c = buffer[i];
		if (!(c >= 'A' && c <= 'Z') &&
			!(c >= 'a' && c <= 'z') &&
			!(c >= '0' && c <= '9')) {
			for (int j = i; j < length; j++) {
				buffer[j] = buffer[j + 1];
			}
			length--;
		} else if (lower && c >= 'A' && c <= 'Z') {
			buffer[i] = c - 'A' + 'a';
		}
	}
	return length;
}

static int device_id_format(char *buffer, int size,
	const char *prefix, const char *device_id,
	const char *const *parts) {
	int index = 0;
	int length = strlen(prefix);
	if (length < size) {
		memcpy(buffer, prefix, length + 1);
		index += length;
	}
	for (int i = 0; parts[i]; i++) {
		const char *substr = strstr(device_id, parts[i]);
		if (!substr || substr == device_id) {
			continue;
		}
		char c = substr[-1];
		if (c != '\\' && c != '&') {
			continue;
		}
		substr = &substr[strlen(parts[i])];
		length = strlen(substr);
		for (int j = 0; j < length; j++) {
			c = substr[j];
			if (c == '\\' || c == '&') {
				length = j;
				break;
			}
		}
		if (index + 1 + length < size) {
			buffer[index++] = '.';
			memcpy(&buffer[index], substr, length);
			length = id_simplify(&buffer[index], length, true);
			index += length;
			buffer[index] = '\0';
		}
	}
	return index;
}

static int tray_item_extra_network_hwid(struct tray_context_t *context,
	char *buffer, int size) {
	if (!context->tooltip || IS_INTRESOURCE(context->tooltip)) {
		return 0;
	}
	if (context->tooltip == (void *) -1) {
		if (context->allow_send_messages) {
			printf("Request tooltip for connection icon.\n");
			SendMessageA(context->data->window,
				context->data->message,
				context->data->id, WM_MOUSEMOVE);
			context->retry = true;
		}
		return 0;
	}
	int name_length = 0;
	for (int i = 0; context->tooltip[i]; i++) {
		if (context->tooltip[i] == L'\n') {
			name_length = i;
			break;
		}
	}
	if (name_length <= 0) {
		name_length = wcslen(context->tooltip);
	}
	if (name_length >= MAX_PATH) {
		return 0;
	}
	ULONG ssize = 0;
	GetAdaptersInfo(NULL, &ssize);
	if (ssize <= 0) {
		return 0;
	}
	IP_ADAPTER_INFO *infos = malloc(ssize);
	if (!infos) {
		return 0;
	}
	DWORD status = GetAdaptersInfo(infos, &ssize);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to enumerate network adapters");
		free(infos);
		return 0;
	}
	HKEY network_hkey = NULL;
	status = RegOpenKeyExA(HKEY_LOCAL_MACHINE,
		"SYSTEM\\CurrentControlSet\\Control\\"
		"Network\\{4D36E972-E325-11CE-BFC1-08002BE10318}",
		0, KEY_ENUMERATE_SUB_KEYS, &network_hkey);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to open network adapters registry key");
		free(infos);
		return 0;
	}
	wchar_t name[name_length + 1];
	char device_id[MAX_PATH] = "\0";
	IP_ADAPTER_INFO *info = infos;
	while (info) {
		ALLOC_FORMAT(key, "%s\\Connection", info->AdapterName);
		info = info->Next;
		HKEY hkey = NULL;
		status = RegOpenKeyExA(network_hkey, key,
			0, KEY_QUERY_VALUE, &hkey);
		if (status != ERROR_SUCCESS) {
			continue;
		}
		DWORD vsize = 2 * (name_length + 1);
		status = RegQueryValueExW(hkey, L"Name",
			NULL, NULL, (void *) name, &vsize);
		name[name_length] = L'\0';
		if (status != ERROR_SUCCESS) {
			RegCloseKey(hkey);
			continue;
		}
		if (wcsstr(context->tooltip, name) == context->tooltip) {
			int length = wcslen(name);
			const wchar_t *c = &context->tooltip[length];
			if (c[0] != L'\n' && c[0] != L'\0' &&
				!(c[0] == L' ' && c[1] == L'(')) {
				RegCloseKey(hkey);
				continue;
			}
		}
		vsize = sizeof(device_id);
		status = RegQueryValueExA(hkey, "PnpInstanceID",
			NULL, NULL, (void *) device_id, &vsize);
		device_id[sizeof(device_id) - 1] = '\0';
		RegCloseKey(hkey);
		break;
	}
	RegCloseKey(network_hkey);
	free(infos);
	if (device_id[0]) {
		if (strstr(device_id, "PCI\\") == device_id) {
			static const char *const parts[] = {
				"VEN_", "DEV_", NULL
			};
			return device_id_format(buffer, size,
				"pci", device_id, parts);
		} else if (strstr(device_id, "USB\\") == device_id) {
			static const char *const parts[] = {
				"VID_", "PID_", "MI_", NULL
			};
			return device_id_format(buffer, size,
				"usb", device_id, parts);
		}
	}
	return 0;
}

static int tray_item_extra_systray_main(struct tray_context_t *context,
	char *buffer, int size) {
	if (context->data->id != context->data->message) {
		return 0;
	}
	int message = context->data->id - WM_USER;
	const char *name = NULL;
	switch (message) {
		case 201: name = "power"; break;
		case 202: name = "hotplug"; break;
		case 203: name = "volume"; break;
		case 204: name = "stickykeys"; break;
		case 205: name = "mousekeys"; break;
		case 206: name = "filterkeys"; break;
	}
	if (name) {
		int length = strlen(name);
		if (length < size) {
			memcpy(buffer, name, length + 1);
			return length;
		}
	} else if (message >= 0) {
		ALLOC_FORMAT(text, "%d", message);
		int length = strlen(text);
		if (length < size) {
			memcpy(buffer, text, length + 1);
			return length;
		}
	}
	return 0;
}

static int tray_item_extra_class(struct tray_context_t *context,
	char *buffer, int size) {
	buffer[0] = '\0';
	GetClassNameA(context->data->window, buffer, size);
	int length = strlen(buffer);
	return id_simplify(buffer, length, false);
}

static int tray_item_add_extra(struct tray_context_t *context,
	char *buffer, int size, int length,
	int (*extra_format)(struct tray_context_t *, char *, int)) {
	int extra_size = size - length - 1;
	if (extra_size > 0) {
		char *extra_buffer = &buffer[length + 1];
		int extra = extra_format(context, extra_buffer, extra_size);
		if (extra > 0) {
			buffer[length] = ':';
			length += extra + 1;
		}
	}
	return length;
}

struct tray_item_full_t {
	char *id;
	int number;
	bool retry;
};

#define TRAY_ITEM_FULL_EMPTY ((struct tray_item_full_t) { 0 })

static struct tray_item_full_t tray_item_from_button(const TBBUTTON *button,
	bool hidden, bool allow_send_messages) {
	if (!button || (!hidden && (button->fsState & TBSTATE_HIDDEN)) ||
		!button->dwData) {
		return TRAY_ITEM_FULL_EMPTY;
	}
	struct tray_context_t context = {
		.data = (void *) button->dwData,
		.tooltip = (void *) button->iString,
		.allow_send_messages = allow_send_messages,
		.retry = false
	};
	if (!context.data->window) {
		return TRAY_ITEM_FULL_EMPTY;
	}
	DWORD pid = 0;
	if (!GetWindowThreadProcessId(context.data->window, &pid)) {
		wperror("Failed to get pid");
	}
	char buffer[2 * MAX_PATH] = "";
	HANDLE process = NULL;
	if (pid) {
		process = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, pid);
		if (!process) {
			wperror("Failed to open process");
		}
	}
	if (process) {
		if (GetProcessImageFileNameA(process,
			buffer, sizeof(buffer)) <= 0) {
			wperror("Failed to get process file name");
		}
		CloseHandle(process);
	}
	int length = strlen(buffer);
	length = id_clear_path(buffer, length);
	length = id_simplify(buffer, length, true);
	bool add_class = true;
	for (int i = 0; config.no_class && config.no_class[i]; i++) {
		if (!strcmp(config.no_class[i], buffer)) {
			add_class = false;
			break;
		}
	}
	if (add_class) {
		const char *check = &buffer[length];
		length = tray_item_add_extra(&context,
			buffer, sizeof(buffer), length,
			tray_item_extra_class);
		if (!*check) {
			fprintf(stderr, "Failed to get class "
				"for \"%s\".\n", buffer);
		}
	} else if (length + 2 <= (int) sizeof(buffer)) {
		buffer[length++] = ':';
		buffer[length++] = TRAY_ID_CHAR_ANY;
		buffer[length] = '\0';
	}
	if (length <= 0) {
		return TRAY_ITEM_FULL_EMPTY;
	}
	if (!strcmp(buffer, "explorer:ConnectionsTray")) {
		length = tray_item_add_extra(&context,
			buffer, sizeof(buffer), length,
			tray_item_extra_network_hwid);
	} else if (!strcmp(buffer, "explorer:SystemTrayMain")) {
		length = tray_item_add_extra(&context,
			buffer, sizeof(buffer), length,
			tray_item_extra_systray_main);
	}
	char *result = malloc(length + 1);
	if (!result) {
		return TRAY_ITEM_FULL_EMPTY;
	}
	memcpy(result, buffer, length + 1);
	return (struct tray_item_full_t) {
		.id = result,
		.number = context.data->id,
		.retry = context.retry
	};
}

struct tray_item_t {
	int index;
	char *id;
	int number;
};

static struct tray_item_t tray_item_from_index(HWND tray_window,
	int index, bool hidden, bool allow_send_messages) {
	TBBUTTON button;
	if (!SendMessageA(tray_window,
		TB_GETBUTTON, index, (LPARAM) &button)) {
		fprintf(stderr, "Failed to get button.\n");
		return (struct tray_item_t) {
			.index = index,
			.id = NULL,
			.number = 0
		};
	}
	struct tray_item_full_t item = tray_item_from_button(&button,
		hidden, allow_send_messages);
	if (item.retry) {
		if (SendMessageA(tray_window,
			TB_GETBUTTON, index, (LPARAM) &button)) {
			struct tray_item_full_t retry_item =
				tray_item_from_button(&button, hidden, false);
			if (retry_item.id) {
				free(item.id);
				item = retry_item;
			}
		} else {
			fprintf(stderr, "Failed to get button.\n");
		}
	}
	return (struct tray_item_t) {
		.index = index,
		.id = item.id,
		.number = item.number
	};
}

static void string_array_free(char **array) {
	for (int i = 0; array && array[i]; i++) {
		free(array[i]);
	}
	free(array);
}

static bool string_array_inc(char ***array_ptr, int current) {
	char **new_array = *array_ptr
		? realloc(*array_ptr, sizeof(**array_ptr) * (current + 2))
		: malloc(sizeof(**array_ptr) * 2);
	if (!new_array) {
		return false;
	}
	if (!*array_ptr) {
		new_array[0] = NULL;
	}
	*array_ptr = new_array;
	return true;
}

static void config_free() {
	string_array_free(config.sorted);
	config.sorted = NULL;
	string_array_free(config.no_class);
	config.no_class = NULL;
}

static void config_load() {
	config_free();
	HKEY hkey;
	DWORD status;
	status = RegOpenKeyExA(HKEY_CURRENT_USER, REGISTRY_KEY,
		0, KEY_QUERY_VALUE, &hkey);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to open registry key");
		return;
	}
	DWORD size;
	status = RegQueryValueExA(hkey, REGISTRY_VALUE_ORDER,
		NULL, NULL, NULL, &size);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to query registry value");
		RegCloseKey(hkey);
		return;
	}
	char *data = malloc(size);
	if (!data) {
		fprintf(stderr, "Failed to allocate memory.\n");
		RegCloseKey(hkey);
		return;
	}
	DWORD type;
	status = RegQueryValueExA(hkey, REGISTRY_VALUE_ORDER,
		NULL, &type, (void *) data, &size);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to query registry value");
		RegCloseKey(hkey);
		free(data);
		return;
	}
	if (type != REG_MULTI_SZ) {
		fprintf(stderr, "Illegal value type.\n");
		RegCloseKey(hkey);
		free(data);
		return;
	}
	int start = 0;
	int index = 0;
	for (int i = 0; i < (int) size; i++) {
		if (data[i] == '\0') {
			if (strlen(&data[start]) > 0) {
				index++;
			}
			start = i + 1;
		}
	}
	start = 0;
	if (index > 0) {
		config.sorted = malloc(sizeof(*config.sorted) * (index + 1));
	}
	if (config.sorted) {
		memset(config.sorted, 0, sizeof(*config.sorted) * (index + 1));
		bool success = true;
		index = 0;
		for (int i = 0; i < (int) size; i++) {
			if (data[i] == '\0') {
				if (strlen(&data[start]) > 0) {
					char *id = strdup(&data[start]);
					if (!id) {
						success = false;
						break;
					}
					config.sorted[index++] = id;
				}
				start = i + 1;
			}
		}
		if (!success) {
			config_free();
		}
	}
	RegCloseKey(hkey);
	free(data);
	printf("Config loaded.\n");
	int no_class_count = 0;
	for (int i = 0; config.sorted && config.sorted[i]; i++) {
		printf("Index %d for \"%s\".\n", i, config.sorted[i]);
		int divider = -1;
		for (int j = 0; config.sorted[i][j]; j++) {
			if (config.sorted[i][j] == ':') {
				divider = j;
				break;
			}
		}
		char *id = NULL;
		if (divider && config.sorted[i][divider + 1] ==
			TRAY_ID_CHAR_ANY && !config.sorted[i][divider + 2]) {
			id = malloc(divider);
			if (id) {
				strncpy(id, config.sorted[i], divider);
				id[divider] = '\0';
			}
		}
		if (id && string_array_inc(&config.no_class, no_class_count)) {
			config.no_class[no_class_count++] = id;
		} else {
			free(id);
		}
	}
	if (no_class_count > 0) {
		config.no_class[no_class_count] = NULL;
	}
	for (int i = 0; config.no_class && config.no_class[i]; i++) {
		printf("Ignore class for \"%s\".\n", config.no_class[i]);
	}
}

static void config_save() {
	int length = 0;
	for (int i = 0; config.sorted && config.sorted[i]; i++) {
		length += strlen(config.sorted[i]) + 1;
	}
	length++;
	char *data = malloc(length);
	if (!data) {
		return;
	}
	int index = 0;
	for (int i = 0; config.sorted && config.sorted[i]; i++) {
		int ilength = strlen(config.sorted[i]) + 1;
		memcpy(&data[index], config.sorted[i], ilength);
		index += ilength;
	}
	data[index++] = '\0';
	HKEY hkey;
	int status;
	status = RegCreateKeyExA(HKEY_CURRENT_USER, REGISTRY_KEY,
		0, NULL, 0, KEY_ALL_ACCESS, NULL, &hkey, NULL);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to open registry key");
		free(data);
		return;
	}
	status = RegSetValueExA(hkey, REGISTRY_VALUE_ORDER,
		0, REG_MULTI_SZ, (void *) data, length);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to set registry value");
	}
	RegCloseKey(hkey);
	free(data);
}

static int sorted_index_of(const char *id) {
	if (!id) {
		return -1;
	}
	for (int i = 0; config.sorted && config.sorted[i]; i++) {
		if (!strcmp(config.sorted[i], id)) {
			return i;
		}
	}
	return -1;
}

static bool ignore_move_button_event = false;

#define INT_COMPARE(a, b) ((a) < (b) ? -1 : (a) > (b) ? 1 : 0)

static int tray_item_sort_compare(const void *a, const void *b) {
	const struct tray_item_t *ta = a;
	const struct tray_item_t *tb = b;
	int ia = sorted_index_of(ta->id);
	int ib = sorted_index_of(tb->id);
	int result = ia < 0 ? -1 : ib < 0 ? 1 : 0;
	result = result != 0 ? result : INT_COMPARE(ia, ib);
	result = result != 0 ? result : INT_COMPARE(ta->number, tb->number);
	return result;
}

static void tray_sort(HWND tray_window, bool print) {
	int count = (int) SendMessageA(tray_window, TB_BUTTONCOUNT, 0, 0);
	if (count <= 0) {
		return;
	}
	struct tray_item_t items[count];
	for (int i = 0; i < count; i++) {
		items[i] = tray_item_from_index(tray_window, i, false, true);
	}
	qsort(items, count, sizeof(*items), tray_item_sort_compare);
	ignore_move_button_event = true;
	for (int i = 0; i < count; i++) {
		struct tray_item_t *item = &items[i];
		if (item->index != i) {
			int from = item->index;
			SendMessageA(tray_window,
				TB_MOVEBUTTON, item->index, i);
			for (int j = i + 1; j < count; j++) {
				item = &items[j];
				if (item->index < from) {
					item->index++;
				}
			}
		}
	}
	ignore_move_button_event = false;
	for (int i = 0; i < count; i++) {
		struct tray_item_t *item = &items[i];
		if (print) {
			printf("Set index %d for \"%s\" [id %d].\n",
				i, item->id ? item->id : "null",
				item->number);
		}
		free(item->id);
	}
}

static void sorted_set_before(const char *what, const char *before) {
	if (!what) {
		return;
	}
	int what_index = sorted_index_of(what);
	int before_index = sorted_index_of(before);
	if (what_index >= 0 && (what_index == before_index - 1 ||
		what_index == before_index)) {
		return;
	}
	int count;
	for (count = 0; config.sorted && config.sorted[count]; count++);
	if (!before) {
		before_index = count;
	} else if (before_index < 0) {
		char *before_copy = strdup(before);
		if (!before_copy || !string_array_inc(&config.sorted, count++)) {
			free(before_copy);
			return;
		}
		for (int i = count; i > 0; i--) {
			config.sorted[i] = config.sorted[i - 1];
		}
		config.sorted[0] = before_copy;
		before_index = 0;
		if (what_index >= 0) {
			what_index++;
		}
	}
	if (what_index >= 0) {
		char *what_copy = config.sorted[what_index];
		if (what_index >= before_index) {
			for (int i = what_index; i > before_index; i--) {
				config.sorted[i] = config.sorted[i - 1];
			}
			config.sorted[before_index] = what_copy;
		} else {
			for (int i = what_index; i < before_index - 1; i++) {
				config.sorted[i] = config.sorted[i + 1];
			}
			config.sorted[before_index - 1] = what_copy;
		}
	} else {
		char *what_copy = strdup(what);
		if (!what_copy || !string_array_inc(&config.sorted, count++)) {
			free(what_copy);
			return;
		}
		for (int i = count; i > before_index; i--) {
			config.sorted[i] = config.sorted[i - 1];
		}
		config.sorted[before_index] = what_copy;
	}
}

static void tray_move(HWND tray_window, int from, int to) {
	if (from == to) {
		return;
	}
	int count = (int) SendMessageA(tray_window, TB_BUTTONCOUNT, 0, 0);
	if (count <= 0) {
		return;
	}
	char *what = tray_item_from_index(tray_window, from, false, false).id;
	if (!what) {
		return;
	}
	char *next = NULL;
	for (int i = count - 1; i >= 0; i--) {
		char *id = i == from ? what :
			tray_item_from_index(tray_window, i, false, false).id;
		if (!id) {
			continue;
		}
		if (sorted_index_of(id) < 0) {
			sorted_set_before(id, next);
		}
		if (next != what) {
			free(next);
		}
		next = id;
	}
	if (next != what) {
		free(next);
	}
	char *before = NULL;
	int before_index = to >= from ? to + 1 : to;
	while (!before && before_index < count) {
		before = tray_item_from_index(tray_window,
			before_index++, false, false).id;
	}
	sorted_set_before(what, before);
	free(before);
	free(what);
	config_save();
}

static HWND get_tray_window() {
	HWND window = FindWindowA("Shell_TrayWnd", NULL);
	if (window) {
		window = FindWindowExA(window, NULL, "TrayNotifyWnd", NULL);
	}
	if (window) {
		window = FindWindowExA(window, NULL, "SysPager", NULL);
	}
	if (window) {
		window = FindWindowExA(window, NULL, "ToolbarWindow32", NULL);
	}
	return window;
}

static void tray_item_print(const char *message,
	const struct tray_item_t *item) {
	printf("%s \"%s\".\n", message, item->id ? item->id : "null");
}

static void tray_item_print_all(const char *message,
	const TBBUTTON *buttons, int count) {
	for (int i = 0; i < count; i++) {
		struct tray_item_full_t item =
			tray_item_from_button(&buttons[i], true, false);
		tray_item_print(message, &(struct tray_item_t) {
			.index = -1,
			.id = item.id,
			.number = item.number
		});
		free(item.id);
	}
}

static bool toolbar_is_show_button_event(HWND tray_window,
	int message, WPARAM wparam, LPARAM lparam) {
	int mask = 0;
	int state = 0;
	if (message == TB_SETBUTTONINFOA) {
		const TBBUTTONINFOA *info = (void *) lparam;
		if (info->cbSize >= offsetof(TBBUTTONINFO, fsState) +
			sizeof(info->fsState)) {
			mask = info->dwMask;
			state = info->fsState;
		}
	} else if (message == TB_SETBUTTONINFOW) {
		const TBBUTTONINFOW *info = (void *) lparam;
		if (info->cbSize >= offsetof(TBBUTTONINFO, fsState) +
			sizeof(info->fsState)) {
			mask = info->dwMask;
			state = info->fsState;
		}
	}
	if ((mask & TBIF_BYINDEX) && (mask & TBIF_STATE) &&
		!(state & TBSTATE_HIDDEN)) {
		TBBUTTONINFOA info = {
			.cbSize = sizeof(info),
			.dwMask = TBIF_STATE | TBIF_BYINDEX
		};
		SendMessageA(tray_window, TB_GETBUTTONINFOA,
			wparam, (LPARAM) &info);
		return !!(info.fsState & TBSTATE_HIDDEN);
	}
	return false;
}

enum {
	XX_TRAY_SORT = (WM_USER << 1) + 1
};

enum process_type_t {
	PROCESS_TYPE_INVALID,
	PROCESS_TYPE_MAIN,
	PROCESS_TYPE_EXPLORER
};

static enum process_type_t process_type = PROCESS_TYPE_INVALID;
static HWND tray_window = NULL;
static bool has_console = false;

static LRESULT CALLBACK hook_msg(int code, WPARAM wparam, LPARAM lparam) {
	if (process_type == PROCESS_TYPE_EXPLORER &&
		code == HC_ACTION && wparam == PM_REMOVE) {
		MSG *msg = (void *) lparam;
		bool tray_sort_event = msg->message == XX_TRAY_SORT;
		if (tray_sort_event && msg->hwnd == tray_window) {
			MSG check_msg;
			if (!PeekMessageA(&check_msg, tray_window,
				XX_TRAY_SORT, XX_TRAY_SORT, PM_NOREMOVE)) {
				printf("Sort tray.\n");
				tray_sort(tray_window, has_console);
			}
		}
	}
	return CallNextHookEx(NULL, code, wparam, lparam);
}

static LRESULT CALLBACK hook_call(int code, WPARAM wparam, LPARAM lparam) {
	if (process_type == PROCESS_TYPE_EXPLORER && code == HC_ACTION) {
		CWPSTRUCT *call = (void *) lparam;
		bool add_buttons_event = call->message == TB_ADDBUTTONSA ||
			call->message == TB_ADDBUTTONSW;
		bool insert_button_event = call->message == TB_INSERTBUTTONA ||
			call->message == TB_INSERTBUTTONW;
		bool move_button_event = call->message == TB_MOVEBUTTON &&
			!ignore_move_button_event;
		bool show_button_event = call->hwnd == tray_window &&
			toolbar_is_show_button_event(tray_window,
				call->message, call->wParam, call->lParam);
		static bool sort_initial = true;
		bool sort_tray = sort_initial;
		if ((add_buttons_event || insert_button_event ||
			move_button_event || show_button_event) &&
			call->hwnd == tray_window) {
			sort_tray = true;
			if (add_buttons_event && has_console) {
				const TBBUTTON *buttons = (void *) call->lParam;
				tray_item_print_all("Add button",
					buttons, call->wParam);
			}
			if (insert_button_event && has_console) {
				const TBBUTTON *button = (void *) call->lParam;
				tray_item_print_all("Insert button",
					button, 1);
			}
			const char *message = NULL;
			if (move_button_event) {
				message = "Move button";
			} else if (show_button_event) {
				message = "Show button";
			}
			if (message && has_console) {
				struct tray_item_t item;
				item = tray_item_from_index(tray_window,
					call->wParam, true, false);
				tray_item_print(message, &item);
				free(item.id);
			}
			if (move_button_event) {
				tray_move(tray_window,
					call->wParam, call->lParam);
			}
		}
		if (sort_tray) {
			sort_initial = false;
			PostMessage(tray_window, XX_TRAY_SORT, 0, 0);
		}
	}
	return CallNextHookEx(NULL, code, wparam, lparam);
}

__attribute__((section(".shared"), shared))
static struct {
	DWORD main_pid;
	bool attach_console;
} shared = {
	.main_pid = 0,
	.attach_console = false
};

static BOOL WINAPI console_handler_ignore(UNUSED DWORD ctrl_type) {
	return ctrl_type == CTRL_C_EVENT || ctrl_type == CTRL_CLOSE_EVENT;
}

BOOL WINAPI DllMain(UNUSED HINSTANCE instance,
	DWORD reason, UNUSED LPVOID reserved) {
	if (reason == DLL_PROCESS_ATTACH) {
		DWORD pid = GetCurrentProcessId();
		tray_window = get_tray_window();
		if (tray_window) {
			DWORD tray_pid;
			if (GetWindowThreadProcessId(tray_window,
				&tray_pid) && tray_pid == pid) {
				process_type = PROCESS_TYPE_EXPLORER;
			}
		}
		if (process_type != PROCESS_TYPE_EXPLORER &&
			!shared.main_pid) {
			shared.main_pid = pid;
			process_type = PROCESS_TYPE_MAIN;
		}
	}
	if (process_type != PROCESS_TYPE_EXPLORER) {
		return process_type == PROCESS_TYPE_MAIN;
	}
	if (reason == DLL_PROCESS_ATTACH) {
		has_console = shared.attach_console &&
			AttachConsole(shared.main_pid);
		if (has_console) {
			freopen("CONOUT$", "w", stdout);
			freopen("CONOUT$", "w", stderr);
			SetConsoleCtrlHandler(console_handler_ignore, TRUE);
		}
		printf("Attach DLL.\n");
		config_load();
	} else if (reason == DLL_PROCESS_DETACH) {
		printf("Detach DLL.\n");
		config_free();
		if (has_console) {
			SetConsoleCtrlHandler(console_handler_ignore, FALSE);
			FreeConsole();
		}
	}
	return TRUE;
}

static bool check_sleep(int ms) {
	MSG msg;
	bool value;
	DWORD error = ERROR_SUCCESS;
	UINT_PTR timer = SetTimer(NULL, 0, ms, NULL);
	while (true) {
		int result = GetMessageA(&msg, NULL, 0, 0);
		if (result == -1) {
			value = false;
			error = GetLastError();
			break;
		} else if (result == 0) {
			value = false;
			break;
		} else if (msg.message == WM_TIMER &&
			(UINT_PTR) msg.wParam == timer) {
			value = true;
			break;
		}
	}
	KillTimer(NULL, timer);
	if (!value) {
		SetLastError(error);
		wperror("Finish sleep");
	}
	return value;
}

static void inject(bool debug) {
	if (debug) {
		if (AllocConsole()) {
			shared.attach_console = true;
			freopen("CONOUT$", "w", stdout);
			freopen("CONOUT$", "w", stderr);
		} else {
			MessageBoxA(NULL, "Failed to allocate console.",
				NULL, MB_ICONERROR);
			return;
		}
	}
	HMODULE module;
	HOOKPROC hook_msg_fptr = hook_msg;
	HOOKPROC hook_call_fptr = hook_call;
	void *func_ptr = (void *) &hook_call_fptr;
	func_ptr = *(void **) func_ptr;
	if (!GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
		GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
		func_ptr, &module)) {
		wperror("Failed to get module handle");
		if (debug) {
			check_sleep(2000);
		}
		return;
	}
	SetProcessShutdownParameters(0, 0);
	while (true) {
		printf("Waiting for tray window.\n");
		HWND tray_window = NULL;
		bool finish = false;
		while (true) {
			tray_window = get_tray_window();
			if (tray_window) {
				break;
			} else if (!check_sleep(1000)) {
				finish = true;
				break;
			}
		}
		if (finish) {
			break;
		}
		DWORD tid = GetWindowThreadProcessId(tray_window, NULL);
		HANDLE thread = tid ? OpenThread(SYNCHRONIZE,
			FALSE, tid) : NULL;
		if (thread) {
			HHOOK handle_msg = SetWindowsHookEx(WH_GETMESSAGE,
				hook_msg_fptr, module, tid);
			HHOOK handle_call = SetWindowsHookEx(WH_CALLWNDPROC,
				hook_call_fptr, module, tid);
			if (handle_msg && handle_call) {
				printf("Waiting for thread finish.\n");
				WaitForSingleObject(thread, INFINITE);
				UnhookWindowsHookEx(handle_msg);
				UnhookWindowsHookEx(handle_call);
				CloseHandle(thread);
			} else {
				if (handle_msg) {
					UnhookWindowsHookEx(handle_msg);
				}
				if (handle_call) {
					UnhookWindowsHookEx(handle_call);
				}
				wperror("Failed to set hook");
				CloseHandle(thread);
				if (!check_sleep(1000)) {
					break;
				}
			}
		} else {
			wperror("Failed to open thread");
			if (!check_sleep(1000)) {
				break;
			}
		}
	}
}

__declspec(dllexport) void run() {
	inject(false);
}

__declspec(dllexport) void debug() {
	inject(true);
}
